// Copyright (c) 1999-2010  Pavel Rychly

#ifndef FINLIB_REGEXPLEX_HH
#define FINLIB_REGEXPLEX_HH

#include <finlib/fsop.hh>
#include <finlib/regpref.hh>
#include <finlib/generator.hh>


template <class Revidx, class Lexicon>
FastStream *regexp2poss (Revidx &rev, Lexicon &lex, const char *pattern, 
                         const char *locale = NULL, const char *encoding = NULL, 
                         bool ignorecase = false)
{
    regexp_pattern pat (pattern, locale, encoding, ignorecase);
    if (pat.no_meta_chars()) {
        int id = lex.str2id (pattern);
        if (id < 0)
            return new EmptyStream();
        else
            return rev.id2poss (id);
    }

    Generator<int> *ids = lex.pref2ids (pat.prefix());
    if (ids->end())
        return new EmptyStream();

    if (pat.compile())
        return new EmptyStream();

    pss_vector *fsv = new pss_vector;
    fsv->reserve(32);
    while (!ids->end()) {
        int id = ids->next();
        if (pat.match (lex.id2str (id))) {
            FastStream *s = rev.id2poss (id);
            fsv->push_back(pos_stream_pair (s->peek(), s));
        }
    }
    delete ids;

    switch (fsv->size()) {
    case 0:
        delete fsv;
        return new EmptyStream();
    case 1: {
        FastStream *f = fsv->front().second;
        delete fsv;
        return f;
    }
    default:
        return new QOrVNode (fsv);
    }
}

template <class Revidx, class Lexicon>
FastStream *compare2poss (Revidx &rev, Lexicon &lex, const char *pattern,
                         int order, bool ignorecase = false)
{
    pss_vector *fsv = new pss_vector;
    fsv->reserve(32);
    for (int id = 0; id < lex.size(); id++) {
        const char *str = lex.id2str (id);
        int cmp = strverscmp(str, pattern);
        if ((order < 0 && cmp <= 0) || (order > 0 && cmp >= 0)) {
            FastStream *s = rev.id2poss (id);
            fsv->push_back(pos_stream_pair (s->peek(), s));
        }
    }

    switch (fsv->size()) {
    case 0:
        delete fsv;
        return new EmptyStream();
    case 1: {
        FastStream *f = fsv->front().second;
        delete fsv;
        return f;
    }
    default:
        return new QOrVNode (fsv);
    }
}

template <class Lexicon>
class regexp2idsStream : public Generator<int>
{
protected:
    Lexicon &lex;
    Generator<int> *lexit;
    regexp_pattern *pat;
    int currid;
    bool finished;
    void locate () {
        finished = true;
        while (!lexit->end()) {
            currid = lexit->next();
            if (pat->match (lex.id2str (currid))) {
                finished = false;
                break;
            }
        }
    }
public:
    regexp2idsStream (Lexicon &l, Generator<int> *li, regexp_pattern *p)
        : lex (l), lexit (li), pat (p), finished (false) {
        locate();
    }
    virtual int next() {
        int ret = currid;
        locate(); 
        return ret;
    }
    virtual bool end() {return finished;}
    virtual int size() {return finished ? 0 : lexit->size() +1;}
    virtual ~regexp2idsStream() {delete lexit; delete pat;}
};


template <class Lexicon>
Generator<int> *regexp2ids (Lexicon &lex, const char *pattern,
        const char *locale = NULL, const char *encoding = NULL,
        bool ignorecase = false)
{
    regexp_pattern *pat = new regexp_pattern (pattern, locale, encoding, ignorecase);
    if (pat->no_meta_chars() && !ignorecase) {
        int id = lex.str2id (pattern);
        delete pat;
        if (id < 0)
            return new EmptyGenerator<int>();
        else
            return new SequenceGenerator<int> (id, id);
    }

    Generator<int> *ids;
    if (ignorecase)
        ids = new SequenceGenerator<int> (0, lex.size() -1);
    else
        ids = lex.pref2ids (pat->prefix());
    if (ids->end())
        return new EmptyGenerator<int>();
    if (pat->compile()) {
        delete pat;
        return new EmptyGenerator<int>();
    }
    return new regexp2idsStream<Lexicon> (lex, ids, pat);
}

#endif

// vim: ts=4 sw=4 sta et sts=4 si cindent tw=80:

